using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;

// CustomUnityAsset Color Changer with tweaks by Blazedust. Original can be found here:
// https://www.reddit.com/r/VAMscenes/comments/a2mwi4/script_customunityasset_color_change_3axiz_scaling/

namespace DillDoe
{
    public class ColorScale : MVRScript
    {
#region PluginInfo    
        public string pluginAuthor = "DillDoe with modifications by Blazedust";
        public string pluginName = "ColorScaleB";
        public string pluginVersion = "1.3";
        public string pluginDate = "01/01/2019";
        public string pluginDescription = @"
        This plugin allows you to change the material color and size of the custom assets.  

        You are free to edit/change everything below this, please do not delete the PluginInfo. Append it if you modify the script 
		
		Edits by Blazedust to make all Renderer components in the GameObject to be affected without having to rely on an internal list in the plugin.";
		
#endregion


#region Vars
        protected JSONStorableColor dcolor;
        protected JSONStorableColor scolor;
        protected JSONStorableFloat gfloat;
        protected JSONStorableFloat mfloat;
        protected JSONStorableFloat afloat;
        protected JSONStorableFloat jfloatX;
        protected JSONStorableFloat jfloatY;
        protected JSONStorableFloat jfloatZ;

		protected JSONStorableFloat dcfloatH;
		protected JSONStorableFloat dcfloatS;
		protected JSONStorableFloat dcfloatV;

		protected JSONStorableFloat scfloatH;
		protected JSONStorableFloat scfloatS;
		protected JSONStorableFloat scfloatV;

		protected JSONStorableFloat theChild;
		protected JSONStorableFloat jfloatAll;
		protected JSONStorableBool scaleComponents;
		// protected JSONStorableString jstring;
		protected UIDynamicSlider dslider;
        // protected UIDynamicTextField dtext;
		protected int loadWaiter = 0;
		protected int mIndex;
		protected JSONStorableBool[] affectToggles;
		protected UIDynamicToggle[] affectTogglesGui;

		private Dictionary<GameObject, Vector3> originalMainScales;
		
		// 200 is plenty for smaller and bigger objects!
		const int MAX_COMPONENT_TOGGLES = 200;
		
		protected List<GOComp> comps;
		
		public class GOComp {
			
			public GameObject go;
			public Vector3 origScale;
			public int componentIndex;

			public JSONStorableBool affectToggle;
			
			public GOComp(GameObject g, int index)
			{
				go = g;
				componentIndex = index;
				if (go.transform.parent != null)
				{
					origScale = go.transform.parent.transform.localScale;
				} 
				else
                {
					origScale = new Vector3(1f, 1f, 1f);
                }
				affectToggle = null;
			}
		}
        
#endregion

		public bool ComponentIncluded(GOComp c)
		{
			if (c.affectToggle != null && !c.affectToggle.val) {
				return false;
			}
			return true;
		}

#region Init
        public override void Init()
        {
            try
            {				
				comps = new List<GOComp>();
				originalMainScales = new Dictionary<GameObject, Vector3>();

				// ************************************  Left Side  *************************************************************

				// Diffuse Color
				HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                dcolor = new JSONStorableColor("DiffuseColor", hsvc, DiffuseColorCallback);
                // RegisterColor(dcolor);
                CreateColorPicker(dcolor);
				
                // Specular Color
                HSVColor hsvc2 = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                scolor = new JSONStorableColor("SpecularColor", hsvc, SpecColorCallback);
                // RegisterColor(scolor);
                CreateColorPicker(scolor);

				dcfloatH = new JSONStorableFloat("DiffuseColorH", 1f, DiffuseColorHCallback, 0f, 1f);
				RegisterFloat(dcfloatH);
				dcfloatS = new JSONStorableFloat("DiffuseColorS", 1f, DiffuseColorSCallback, 0f, 1f);
				RegisterFloat(dcfloatS);
				dcfloatV = new JSONStorableFloat("DiffuseColorV", 1f, DiffuseColorVCallback, 0f, 1f);
				RegisterFloat(dcfloatV);

				scfloatH = new JSONStorableFloat("SpecularColorH", 1f, SpecularColorHCallback, 0f, 1f);
				RegisterFloat(scfloatH);
				scfloatS = new JSONStorableFloat("SpecularColorS", 1f, SpecularColorSCallback, 0f, 1f);
				RegisterFloat(scfloatS);
				scfloatV = new JSONStorableFloat("SpecularColorV", 1f, SpecularColorVCallback, 0f, 1f);
				RegisterFloat(scfloatV);

				//  Spec
				gfloat = new JSONStorableFloat("Glossiness", 0.2f, GlossyCallback, 0f, 1f, true);
                RegisterFloat(gfloat);
                dslider = CreateSlider(gfloat);

                //  Metallic
                mfloat = new JSONStorableFloat("Metallic", 0f, MetallicCallback, 0f, 1f, true);
                RegisterFloat(mfloat);
                dslider = CreateSlider(mfloat);

                //  Alpha
                afloat = new JSONStorableFloat("Alpha", 1f, AlphaCallback, 0f, 1f, true);
                RegisterFloat(afloat);
                dslider = CreateSlider(afloat);


				// ************************************  Right Side  *************************************************************

				// Scale X
				JSONStorableFloat jfloatAll = new JSONStorableFloat("ScaleAll", 1f, ScaleAllCallback, -1f, 2f, false, true);
				CreateSlider(jfloatAll, true);

				// Scale X
				jfloatX = new JSONStorableFloat("ScaleX", 1f, ScaleX, -1f, 2f, false, true);
                //jfloatX.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(jfloatX);
                CreateSlider(jfloatX, true);

                // Scale Y
                jfloatY = new JSONStorableFloat("ScaleY", 1f, ScaleY, -1f, 2f, false, true);
                RegisterFloat(jfloatY);
                CreateSlider(jfloatY, true);

                // Scale Z
                jfloatZ = new JSONStorableFloat("ScaleZ", 1f, ScaleZ, -1f, 2f, false, true);
                RegisterFloat(jfloatZ);
                CreateSlider(jfloatZ, true);

				scaleComponents = new JSONStorableBool("ScaleComponents", false, ScaleComponentsCallback);
				RegisterBool(scaleComponents);
				UIDynamicToggle toggle = CreateToggle(scaleComponents, true);
				toggle.label = "Scale Components (On)\nWhole Object (Off)";
				toggle.height = 80;

				/*UIDynamic spacer = CreateSpacer(true);
                spacer.height = 240f;
                */

				// jstring = new JSONStorableString("Childs", "");
				// dtext = CreateTextField(jstring,true);
				// if (go != null)
				// {
				//     FindChild();
				// }

				theChild = new JSONStorableFloat("Material", 1f, ChildCallback, 1f, 10);
                RegisterFloat(theChild);
                dslider = CreateSlider(theChild, true);
                dslider.slider.wholeNumbers = true;
				
				// Pre-create toggle-boxes over which ones should affect and not.
				affectToggles = new JSONStorableBool[MAX_COMPONENT_TOGGLES];
				affectTogglesGui = new UIDynamicToggle[affectToggles.Length];
				for (int i = 0; i < affectToggles.Length; i ++) 
				{
					affectToggles[i] = new JSONStorableBool("component_" + i.ToString(), false);
					RegisterBool(affectToggles[i]);
				}

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        #endregion

		void OnDestroy()
        {
			RemoveColorPicker(dcolor);
			RemoveColorPicker(scolor);
			RemoveSlider(jfloatAll);
			RemoveSlider(gfloat);
			RemoveSlider(mfloat);
			RemoveSlider(afloat);
			RemoveSlider(jfloatX);
			RemoveSlider(jfloatY);
			RemoveSlider(jfloatZ);
			RemoveSlider(theChild);
			RemoveSlider(dslider);
			RemoveToggle(scaleComponents);

			for (int i = 0; i < affectToggles.Length; i ++)
            {
				if (affectToggles[i] != null)
				{
					RemoveToggle(affectToggles[i]);
				}
			}
		}

			private void InitComponentsGUI() 
		{
			
			UIDynamicButton refreshAll = CreateButton("Refresh selected components", true);
            refreshAll.button.onClick.AddListener(() =>
            {
				try {
					LoadVal();
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });
			refreshAll.buttonColor = new Color(0.7f, 0.7f, 0.9f);
			
			// Create a "Select all" and "Deselect all" button.
			UIDynamicButton selectAll = CreateButton("Select All Components to affect", true);
            selectAll.button.onClick.AddListener(() =>
            {
				try {
					SetAllComponents(true);
					LoadVal();
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });
			selectAll.buttonColor = new Color(0.7f, 0.9f, 0.7f);
			
			UIDynamicButton deselectAll = CreateButton("Deselect All Components", true);
            deselectAll.button.onClick.AddListener(() =>
            {
				try {
					SetAllComponents(false);
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });
			deselectAll.buttonColor = new Color(0.9f, 0.7f, 0.7f);
			
			int max = Math.Min(comps.Count, affectToggles.Length);
			for (int i = 0; i < max; i ++) 
			{
				affectTogglesGui[i] = CreateToggle(affectToggles[i], true);
				affectTogglesGui[i].label = i.ToString() + ":" + comps[i].go.name;
				affectTogglesGui[i].backgroundColor = Color.yellow;
				comps[i].affectToggle = affectToggles[i]; 
			}
		}

		private void SetAllComponents(bool val)
		{
			for (int i = 0; i < affectToggles.Length; i ++) 
			{
				if (affectToggles[i] != null) 
				{
					affectToggles[i].val = val;
				} else {
					// reached end of toggles
					break;
				}
			}
		}

#region Update
        void Update()
        {
			if (loadWaiter == 200 || SuperController.singleton.isLoading) 
			{
				return;
			}
            try
            {
				//  There seems to be a slight delay that prevents finding gameobject on load,
				//  so wait for 200 update cycles after main loading is over before trying.

				loadWaiter += 1;
				if (loadWaiter == 200) 
				{
					comps = new List<GOComp>();
					
					Atom atom = GetContainingAtom();
					
					int componentIndex = 0;
					foreach (Transform t in atom.onToggleObjects) 
					{
						if (t != null && t.gameObject != null) 
						{
							originalMainScales[t.gameObject] = t.gameObject.transform.localScale;

							Component[] components = t.gameObject.GetComponentsInChildren(typeof(Renderer)); // SkinnedMeshRenderer, MeshRenderer;
							//Component[] components = t.gameObject.GetComponentsInChildren(typeof(MeshRenderer)); // SkinnedMeshRenderer, MeshRenderer;
							//Component[] components = t.gameObject.GetComponentsInChildren(typeof(SkinnedMeshRenderer)); // SkinnedMeshRenderer, MeshRenderer;
							if (components != null && components.Length > 0) 
							{
								for (int k = 0; k < components.Length; k ++) 
								{
									// SuperController.LogMessage("component " + componentIndex.ToString() + ":" + components[k].name + ", " + components[k].GetType().ToString());	
									comps.Add(new GOComp(components[k].gameObject, componentIndex));
									componentIndex ++;
								}
							}

						}
					}
					
					if (comps.Count > 0)
					{
						// SuperController.LogMessage("ColorScale: " + comps.Count.ToString() + " components to load");
						InitComponentsGUI();
						LoadVal();
					} 
					else 
					{
						SuperController.LogMessage("ColorScale: Nothing to load");
					}
				}
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        #endregion

#region Load Values
        //  Load all the saved values from plugin UI back into Asset
        void LoadVal()
        {
            try
            {
                ScaleX(jfloatX);
                ScaleY(jfloatY);
                ScaleZ(jfloatZ);

                // FindChild();
                ChildCallback(theChild);

                DiffuseColorCallback(dcolor);
                SpecColorCallback(scolor);
                GlossyCallback(gfloat);
                MetallicCallback(mfloat);
                AlphaCallback(afloat);

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
#endregion

#region UI Functions

		// private void Scale(float value)
        //{
			// // To scale whole object:
			//foreach (Transform t in GetContainingAtom().onToggleObjects)
			//{
			//    if (t != null && t.gameObject != null)
			//    {
			//        t.gameObject.GetComponent<Transform>().localScale = new Vector3(value, value, value);
			//    }
			//}
		//}

		protected void ScaleComponentsCallback(JSONStorableBool jb)
        {
			ScaleX(jfloatX);
			ScaleY(jfloatY);
			ScaleZ(jfloatZ);
		}
		
		protected void ScaleAllCallback(JSONStorableFloat jAll)
        {
			jfloatX.val = jAll.val;
			jfloatY.val = jAll.val;
			jfloatZ.val = jAll.val;
		}

		protected void ScaleC(Func<Vector3, Vector3, Vector3> compScale)
		{
			Atom atom = GetContainingAtom();
			if (scaleComponents.val)
			{
				// Restore original
				foreach (Transform t in atom.onToggleObjects)
				{
					if (t != null && t.gameObject != null)
					{
						Vector3 originalScale = Vector3.zero;
						if (originalMainScales.TryGetValue(t.gameObject, out originalScale))
						{
							Transform trans = t.gameObject.transform;
							trans.localScale = originalScale;
						}
					}
				}

				// Apply new 
				comps.ForEach((x) =>
				{
					if (ComponentIncluded(x))
					{
						if (x.go.transform != null && x.go.transform.parent != null)
						{
							Transform trans = x.go.transform.parent.transform;
							if (trans != null)
							{
								trans.localScale = compScale(x.origScale, trans.localScale);
							}
						}
					}
				});
			}
			else
			{
				// Restore original
				comps.ForEach((x) =>
				{
					// if (ComponentIncluded(x))
					{
						if (x.go.transform != null && x.go.transform.parent != null)
						{
							Transform trans = x.go.transform.parent.transform;
							if (trans != null)
							{
								trans.localScale = x.origScale;
							}
						}
					}
				});

				// Apply new 
				foreach (Transform t in atom.onToggleObjects)
				{
					if (t != null && t.gameObject != null)
					{
						Vector3 originalScale = Vector3.zero;
						if (originalMainScales.TryGetValue(t.gameObject, out originalScale))
						{
							Transform trans = t.gameObject.transform;
							trans.localScale = compScale(originalScale, trans.localScale);
						}
					}
				}
			}
		}

		//	Scale		
		protected void ScaleX(JSONStorableFloat jX)
        {
			ScaleC((Vector3 origScale, Vector3 localScale) => { return new Vector3(jX.val * origScale.x, localScale.y, localScale.z); });
        }

        protected void ScaleY(JSONStorableFloat jY)
        {
			ScaleC((Vector3 origScale, Vector3 localScale) => { return new Vector3(localScale.x, jY.val * origScale.y, localScale.z); });
        }

        protected void ScaleZ(JSONStorableFloat jZ)
        {
			ScaleC((Vector3 origScale, Vector3 localScale) => { return new Vector3(localScale.x, localScale.y, jZ.val * origScale.z); });
		}

        //  find all the child object in prefab to change each material color
        // protected void FindChild()
        // {            
        //     if (go != null)
        //     {   // todo: create dynamic array to hold color info for each prefab to save with scene
        //         //Dictionary<string, HSVColor> prefabChild = new Dictionary<string, HSVColor>();
        //         float x = 0f;
        //         foreach (var ob in go.GetComponentsInChildren<Renderer>())
        //         {
        //             x++;
        //             jstring.val += x + ". " + ob.name + "\n";
        //         }             
        //     }
        // }
        // prefab selector
        protected void ChildCallback(JSONStorableFloat jC)
        {
			// SuperController.LogMessage("mIndex pre " + mIndex.ToString());
            mIndex = Math.Max((int)jC.val - 1, 0);
			// SuperController.LogMessage("mIndex post " + mIndex.ToString());
        }

		protected void DiffuseColorHCallback(JSONStorableFloat jf)
        {
			HSVColor hc = dcolor.val;
			hc.H = jf.val;
			dcolor.val = hc;
		}

		protected void DiffuseColorSCallback(JSONStorableFloat jf)
        {
			HSVColor hc = dcolor.val;
			hc.S = jf.val;
			dcolor.val = hc;
		}

		protected void DiffuseColorVCallback(JSONStorableFloat jf)
        {
			HSVColor hc = dcolor.val;
			hc.V = jf.val;
			dcolor.val = hc;
		}

		//  material color
		protected void DiffuseColorCallback(JSONStorableColor dc)
        {
			dcfloatH.val = dc.valNoCallback.H;
			dcfloatS.val = dc.valNoCallback.S;
			dcfloatV.val = dc.valNoCallback.V;
			Color c = Color.HSVToRGB(dc.valNoCallback.H, dc.valNoCallback.S, dc.valNoCallback.V);
            comps.ForEach((x) => {
				if (ComponentIncluded(x)) {
					// foreach(Renderer rend in x.go.GetComponentsInChildren<Renderer>()) {
					// 	rend.material.SetColor("_Color", c);
					Renderer[] rends = x.go.GetComponentsInChildren<Renderer>();
					if (rends != null && mIndex < rends.Length) {
						rends[mIndex].material.SetColor("_Color", c);
					}
				}
            });
        }

		protected void SpecularColorHCallback(JSONStorableFloat jf)
		{
			HSVColor hc = scolor.val;
			hc.H = jf.val;
			scolor.val = hc;
		}

		protected void SpecularColorSCallback(JSONStorableFloat jf)
		{
			HSVColor hc = scolor.val;
			hc.S = jf.val;
			scolor.val = hc;
		}

		protected void SpecularColorVCallback(JSONStorableFloat jf)
		{
			HSVColor hc = scolor.val;
			hc.V = jf.val;
			scolor.val = hc;
		}

		protected void SpecColorCallback(JSONStorableColor sc)
        {
			scfloatH.val = sc.valNoCallback.H;
			scfloatS.val = sc.valNoCallback.S;
			scfloatV.val = sc.valNoCallback.V;
			Color c = Color.HSVToRGB(sc.valNoCallback.H, sc.valNoCallback.S, sc.valNoCallback.V);
            comps.ForEach((x) => {
				if (ComponentIncluded(x)) {
					Renderer[] rends = x.go.GetComponentsInChildren<Renderer>();
					if (rends != null && mIndex < rends.Length) {
						rends[mIndex].material.SetColor("_SpecColor", c);
					}
				}
            });
        }

        protected void GlossyCallback(JSONStorableFloat jf)
        {
            comps.ForEach((x) => {
				if (ComponentIncluded(x)) {
					Renderer[] rends = x.go.GetComponentsInChildren<Renderer>();
					if (rends != null && mIndex < rends.Length) {
						rends[mIndex].material.SetFloat("_Glossiness", jf.val);
					}
				}
            });
        }

        protected void MetallicCallback(JSONStorableFloat jm)
        {
            comps.ForEach((x) => {
				if (ComponentIncluded(x)) {
					Renderer[] rends = x.go.GetComponentsInChildren<Renderer>();
					if (rends != null && mIndex < rends.Length) {
						rends[mIndex].material.SetFloat("_Metallic", jm.val);
					}
				}
            });
        }

        protected void AlphaCallback(JSONStorableFloat ja)
        {
            comps.ForEach((x) => {
				if (ComponentIncluded(x)) {
					Renderer[] rends = x.go.GetComponentsInChildren<Renderer>();
					if (rends != null && mIndex < rends.Length) {
						Color colorA = rends[mIndex].material.color;
						colorA.a = ja.val;
						rends[mIndex].material.SetColor("_Color", colorA);
					}
				}
            });
        }
    }

#endregion
}